#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using DiffSharp.Interop.Float64;

namespace HeuristicLab.Algorithms.DataAnalysis.Experimental {
  [Item("Constant Optimization Evaluator with Constraints", "")]
  [StorableClass]
  public class SymbolicRegressionConstantOptimizationEvaluator : SymbolicRegressionSingleObjectiveEvaluator {
    private const string ConstantOptimizationIterationsParameterName = "ConstantOptimizationIterations";
    private const string ConstantOptimizationImprovementParameterName = "ConstantOptimizationImprovement";
    private const string ConstantOptimizationProbabilityParameterName = "ConstantOptimizationProbability";
    private const string ConstantOptimizationRowsPercentageParameterName = "ConstantOptimizationRowsPercentage";
    private const string UpdateConstantsInTreeParameterName = "UpdateConstantsInSymbolicExpressionTree";
    private const string UpdateVariableWeightsParameterName = "Update Variable Weights";

    public IFixedValueParameter<IntValue> ConstantOptimizationIterationsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[ConstantOptimizationIterationsParameterName]; }
    }
    public IFixedValueParameter<DoubleValue> ConstantOptimizationImprovementParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters[ConstantOptimizationImprovementParameterName]; }
    }
    public IFixedValueParameter<PercentValue> ConstantOptimizationProbabilityParameter {
      get { return (IFixedValueParameter<PercentValue>)Parameters[ConstantOptimizationProbabilityParameterName]; }
    }
    public IFixedValueParameter<PercentValue> ConstantOptimizationRowsPercentageParameter {
      get { return (IFixedValueParameter<PercentValue>)Parameters[ConstantOptimizationRowsPercentageParameterName]; }
    }
    public IFixedValueParameter<BoolValue> UpdateConstantsInTreeParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[UpdateConstantsInTreeParameterName]; }
    }
    public IFixedValueParameter<BoolValue> UpdateVariableWeightsParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[UpdateVariableWeightsParameterName]; }
    }


    public IntValue ConstantOptimizationIterations {
      get { return ConstantOptimizationIterationsParameter.Value; }
    }
    public DoubleValue ConstantOptimizationImprovement {
      get { return ConstantOptimizationImprovementParameter.Value; }
    }
    public PercentValue ConstantOptimizationProbability {
      get { return ConstantOptimizationProbabilityParameter.Value; }
    }
    public PercentValue ConstantOptimizationRowsPercentage {
      get { return ConstantOptimizationRowsPercentageParameter.Value; }
    }
    public bool UpdateConstantsInTree {
      get { return UpdateConstantsInTreeParameter.Value.Value; }
      set { UpdateConstantsInTreeParameter.Value.Value = value; }
    }

    public bool UpdateVariableWeights {
      get { return UpdateVariableWeightsParameter.Value.Value; }
      set { UpdateVariableWeightsParameter.Value.Value = value; }
    }

    public override bool Maximization {
      get { return true; }
    }

    [StorableConstructor]
    protected SymbolicRegressionConstantOptimizationEvaluator(bool deserializing) : base(deserializing) { }
    protected SymbolicRegressionConstantOptimizationEvaluator(SymbolicRegressionConstantOptimizationEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public SymbolicRegressionConstantOptimizationEvaluator()
      : base() {
      Parameters.Add(new FixedValueParameter<IntValue>(ConstantOptimizationIterationsParameterName, "Determines how many iterations should be calculated while optimizing the constant of a symbolic expression tree (0 indicates other or default stopping criterion).", new IntValue(10), true));
      Parameters.Add(new FixedValueParameter<DoubleValue>(ConstantOptimizationImprovementParameterName, "Determines the relative improvement which must be achieved in the constant optimization to continue with it (0 indicates other or default stopping criterion).", new DoubleValue(0), true) { Hidden = true });
      Parameters.Add(new FixedValueParameter<PercentValue>(ConstantOptimizationProbabilityParameterName, "Determines the probability that the constants are optimized", new PercentValue(1), true));
      Parameters.Add(new FixedValueParameter<PercentValue>(ConstantOptimizationRowsPercentageParameterName, "Determines the percentage of the rows which should be used for constant optimization", new PercentValue(1), true));
      Parameters.Add(new FixedValueParameter<BoolValue>(UpdateConstantsInTreeParameterName, "Determines if the constants in the tree should be overwritten by the optimized constants.", new BoolValue(true)) { Hidden = true });
      Parameters.Add(new FixedValueParameter<BoolValue>(UpdateVariableWeightsParameterName, "Determines if the variable weights in the tree should be  optimized.", new BoolValue(true)) { Hidden = true });
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicRegressionConstantOptimizationEvaluator(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(UpdateConstantsInTreeParameterName))
        Parameters.Add(new FixedValueParameter<BoolValue>(UpdateConstantsInTreeParameterName, "Determines if the constants in the tree should be overwritten by the optimized constants.", new BoolValue(true)));
      if (!Parameters.ContainsKey(UpdateVariableWeightsParameterName))
        Parameters.Add(new FixedValueParameter<BoolValue>(UpdateVariableWeightsParameterName, "Determines if the variable weights in the tree should be  optimized.", new BoolValue(true)));
    }

    public override IOperation InstrumentedApply() {
      var solution = SymbolicExpressionTreeParameter.ActualValue;
      double quality;
      if (RandomParameter.ActualValue.NextDouble() < ConstantOptimizationProbability.Value) {
        IEnumerable<int> constantOptimizationRows = GenerateRowsToEvaluate(ConstantOptimizationRowsPercentage.Value);
        quality = OptimizeConstants(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, solution, ProblemDataParameter.ActualValue,
           constantOptimizationRows, ApplyLinearScalingParameter.ActualValue.Value, ConstantOptimizationIterations.Value, updateVariableWeights: UpdateVariableWeights, lowerEstimationLimit: EstimationLimitsParameter.ActualValue.Lower, upperEstimationLimit: EstimationLimitsParameter.ActualValue.Upper, updateConstantsInTree: UpdateConstantsInTree);

        if (ConstantOptimizationRowsPercentage.Value != RelativeNumberOfEvaluatedSamplesParameter.ActualValue.Value) {
          var evaluationRows = GenerateRowsToEvaluate();
          quality = SymbolicRegressionSingleObjectivePearsonRSquaredEvaluator.Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, solution, EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper, ProblemDataParameter.ActualValue, evaluationRows, ApplyLinearScalingParameter.ActualValue.Value);
        }
      } else {
        var evaluationRows = GenerateRowsToEvaluate();
        quality = SymbolicRegressionSingleObjectivePearsonRSquaredEvaluator.Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, solution, EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper, ProblemDataParameter.ActualValue, evaluationRows, ApplyLinearScalingParameter.ActualValue.Value);
      }
      QualityParameter.ActualValue = new DoubleValue(quality);

      return base.InstrumentedApply();
    }

    public override double Evaluate(IExecutionContext context, ISymbolicExpressionTree tree, IRegressionProblemData problemData, IEnumerable<int> rows) {
      SymbolicDataAnalysisTreeInterpreterParameter.ExecutionContext = context;
      EstimationLimitsParameter.ExecutionContext = context;
      ApplyLinearScalingParameter.ExecutionContext = context;

      // Pearson R evaluator is used on purpose instead of the const-opt evaluator, 
      // because Evaluate() is used to get the quality of evolved models on 
      // different partitions of the dataset (e.g., best validation model)
      double r2 = SymbolicRegressionSingleObjectivePearsonRSquaredEvaluator.Calculate(SymbolicDataAnalysisTreeInterpreterParameter.ActualValue, tree, EstimationLimitsParameter.ActualValue.Lower, EstimationLimitsParameter.ActualValue.Upper, problemData, rows, ApplyLinearScalingParameter.ActualValue.Value);

      SymbolicDataAnalysisTreeInterpreterParameter.ExecutionContext = null;
      EstimationLimitsParameter.ExecutionContext = null;
      ApplyLinearScalingParameter.ExecutionContext = null;

      return r2;
    }

    public static double OptimizeConstants(ISymbolicDataAnalysisExpressionTreeInterpreter interpreter,
      ISymbolicExpressionTree tree, IRegressionProblemData problemData, IEnumerable<int> rows, bool applyLinearScaling,
      int maxIterations, bool updateVariableWeights = true,
      double lowerEstimationLimit = double.MinValue, double upperEstimationLimit = double.MaxValue,
      bool updateConstantsInTree = true) {

      // numeric constants in the tree become variables for constant opt
      // variables in the tree become parameters (fixed values) for constant opt
      // for each parameter (variable in the original tree) we store the 
      // variable name, variable value (for factor vars) and lag as a DataForVariable object.
      // A dictionary is used to find parameters
      double[] initialConstants;
      var parameters = new List<TreeToDiffSharpConverter.DataForVariable>();

      Func<DV, D> func;
      if (!TreeToDiffSharpConverter.TryConvertToDiffSharp(tree, updateVariableWeights, out parameters, out initialConstants,
        out func))
        throw new NotSupportedException("Could not optimize constants of symbolic expression tree due to not supported symbols used in the tree.");
      if (parameters.Count == 0) return 0.0; // gkronber: constant expressions always have a R of 0.0 

      var parameterEntries = parameters.ToArray(); // order of entries must be the same for x

      // extract inital constants
      double[] c = new double[initialConstants.Length + 2];
      double[] s = new double[c.Length];
      {
        c[0] = 1.0;
        c[1] = 0.0;
        Array.Copy(initialConstants, 0, c, 2, initialConstants.Length);

        // s[0] = 1.0;
        // s[1] = 1.0;
        // Array.Copy(initialConstants.Select(ci=>Math.Abs(ci)).ToArray()
        //   , 0, s, 2, initialConstants.Length);
      }
      s = Enumerable.Repeat(0.01, c.Length).ToArray();

      double[] originalConstants = (double[])c.Clone();
      double originalQuality = SymbolicRegressionSingleObjectivePearsonRSquaredEvaluator.Calculate(interpreter, tree, lowerEstimationLimit, upperEstimationLimit, problemData, rows, applyLinearScaling);

      alglib.minnlcstate state;
      alglib.minnlcreport rep;

      IDataset ds = problemData.Dataset;
      double[,] x = new double[ds.Rows, parameters.Count];
      int col = 0;
      int row = 0;
      foreach (var info in parameterEntries) {
        row = 0;
        // copy training rows
        foreach (var r in rows) { 
          if (ds.VariableHasType<double>(info.variableName)) {
            x[row, col] = ds.GetDoubleValue(info.variableName, r + info.lag);
          } else if (ds.VariableHasType<string>(info.variableName)) {
            x[row, col] = ds.GetStringValue(info.variableName, r) == info.variableValue ? 1 : 0;
          } else throw new InvalidProgramException("found a variable of unknown type");

          row++;
        }
        col++;
      }

      var target = problemData.TargetVariable;

      // only count constraints
      var constraintRows = Enumerable.Range(0, problemData.Dataset.Rows).Where(rIdx => !string.IsNullOrEmpty(ds.GetStringValue("f(x) constraint-type", rIdx)));
      double[,] constraints = new double[constraintRows.Count(), parameters.Count + 1]; // +1 for constraint for f(x)
      string[,] comp = new string[constraintRows.Count(), parameters.Count + 1];
      int eqConstraints = 0;
      int ieqConstraints = 0;
      col = 0;
      foreach (var info in parameterEntries) {
        row = 0;
        // find the matching df/dx column
        var colName = string.Format("df/d({0})", info.variableName);
        var compColName = string.Format("df/d({0}) constraint-type", info.variableName);

        if (ds.VariableNames.Contains(colName)) {
          foreach (var r in constraintRows) {
            constraints[row, col] = ds.GetDoubleValue(colName, r);
            comp[row, col] = ds.GetStringValue(compColName, r);

            if (comp[row, col] == "EQ") eqConstraints++;
            else if (comp[row, col] == "LEQ" || comp[row, col] == "GEQ") ieqConstraints++;

            row++;
          }
        }
        col++;
      }

      // f(x) constraint
      row = 0;
      col = constraints.GetLength(1) - 1;
      foreach (var r in constraintRows) {
        constraints[row, col] = ds.GetDoubleValue("f(x)", r);
        comp[row, col] = ds.GetStringValue("f(x) constraint-type", r);
        if (comp[row, col] == "EQ") eqConstraints++;
        else if (comp[row, col] == "LEQ" || comp[row, col] == "GEQ") ieqConstraints++;
        row++;
      }

      double[] y = ds.GetDoubleValues(problemData.TargetVariable, rows).ToArray();

      alglib.ndimensional_jac jac = CreateJac(x, y, ds, func);
      double rho = 1000;
      int outeriters = 3;
      int updateFreq = 10;
      try {
        // alglib.lsfitcreatefg(x, y, c, n, m, k, false, out state);
        // alglib.lsfitsetcond(state, 0.0, maxIterations);
        // //alglib.lsfitsetgradientcheck(state, 0.001);
        // alglib.lsfitfit(state, function_cx_1_func, function_cx_1_grad, null, null);
        // alglib.lsfitresults(state, out retVal, out c, out rep);
        alglib.minnlccreate(c, out state);
        alglib.minnlcsetalgoaul(state, rho, outeriters);
        alglib.minnlcsetcond(state, 0.0, 0.0, 0.0, maxIterations);
        alglib.minnlcsetscale(state, s);
        alglib.minnlcsetprecexactlowrank(state, updateFreq);
        // TODO set constraints;
        alglib.minnlcsetnlc(state, eqConstraints, ieqConstraints);
        alglib.minnlcoptimize(state, jac, null, null);
        alglib.minnlcresults(state, out c, out rep);
      } catch (ArithmeticException) {
        return originalQuality;
      } catch (alglib.alglibexception) {
        return originalQuality;
      }

      //  -7  => constant optimization failed due to wrong gradient
      //  -8  => integrity check failed (e.g. gradient NaN
      if (rep.terminationtype != -7 && rep.terminationtype != -8)
        UpdateConstants(tree, c.Skip(2).ToArray(), updateVariableWeights);
      var quality = SymbolicRegressionSingleObjectivePearsonRSquaredEvaluator.Calculate(interpreter, tree, lowerEstimationLimit, upperEstimationLimit, problemData, rows, applyLinearScaling);

      if (!updateConstantsInTree) UpdateConstants(tree, originalConstants.Skip(2).ToArray(), updateVariableWeights);
      if (
        // originalQuality - quality > 0.001 || 
        double.IsNaN(quality)) {
        UpdateConstants(tree, originalConstants.Skip(2).ToArray(), updateVariableWeights);
        return originalQuality;
      }
      return quality;
    }

    private static void UpdateConstants(ISymbolicExpressionTree tree, double[] constants, bool updateVariableWeights) {
      int i = 0;
      foreach (var node in tree.Root.IterateNodesPrefix().OfType<SymbolicExpressionTreeTerminalNode>()) {
        ConstantTreeNode constantTreeNode = node as ConstantTreeNode;
        VariableTreeNodeBase variableTreeNodeBase = node as VariableTreeNodeBase;
        FactorVariableTreeNode factorVarTreeNode = node as FactorVariableTreeNode;
        if (constantTreeNode != null)
          constantTreeNode.Value = constants[i++];
        else if (updateVariableWeights && variableTreeNodeBase != null)
          variableTreeNodeBase.Weight = constants[i++];
        else if (factorVarTreeNode != null) {
          for (int j = 0; j < factorVarTreeNode.Weights.Length; j++)
            factorVarTreeNode.Weights[j] = constants[i++];
        }
      }
    }

    private static alglib.ndimensional_jac CreateJac(
      double[,] x, // x is larger than y 
      double[] y, // only targets
      // double[,] constraints, // df/d(xi), same order as for x, same number of rows as x less columns
      // string[,] comparison, // {LEQ, GEQ, EQ } same size as constraints
      IDataset ds,
      Func<DV, D> func) {
      return (double[] c, double[] fi, double[,] jac, object o) => {
        // objective function is sum of squared errors
        int nRows = y.Length;
        int nParams = x.GetLength(1);
        // zero fi and jac
        Array.Clear(fi, 0, fi.Length);
        Array.Clear(jac, 0, jac.Length);
        var p = new double[nParams + c.Length];
        Array.Copy(c, 0, p, nParams, c.Length); // copy c to the end of the function parameters vector
        for (int rowIdx = 0; rowIdx < nRows; rowIdx++) {
          // copy x_i to the beginning of the function parameters vector
          for (int cIdx = 0; cIdx < nParams; cIdx++)
            p[cIdx] = x[rowIdx, cIdx];

          double f = (double)func(p);
          double[] g = (double[])AD.Grad(func, p);
          var e = y[rowIdx] - f;
          fi[0] += e * e;
          // update gradient
          for (int colIdx = 0; colIdx < c.Length; colIdx++) {
            jac[0, colIdx] += -2 * e * g[nParams + colIdx]; // skip the elements for the variable values
          }
        }

        int fidx = 1;
        var constraintRows = Enumerable.Range(0, ds.Rows).Where(rIdx => !string.IsNullOrEmpty(ds.GetStringValue("f(x) constraint-type", rIdx))); 

        // eq constraints
        foreach (var rowIdx in constraintRows) {
          foreach (var varName in ds.VariableNames.Where(vn => vn.EndsWith("constraint-type"))) {
            if (ds.GetStringValue(varName, rowIdx) == "EQ") {
              throw new NotSupportedException();
            }
          }
        }
        // ineq constraints
        foreach (var rowIdx in constraintRows) {
XXX 
          for (int colIdx = 0; colIdx < constraints.GetLength(1); colIdx++) {
            // there is a constraint value
            if (!double.IsNaN(constraints[rowIdx, colIdx]) && !string.IsNullOrEmpty(comparison[rowIdx, colIdx])) {
              var factor = (comparison[rowIdx, colIdx] == "LEQ") ? 1.0
              : comparison[rowIdx, colIdx] == "GEQ" ? -1.0 : 0.0;

              // f(x) constraint
              if (colIdx == constraints.GetLength(1) - 1) {
                // copy x_i to the beginning of the function parameters vector
                for (int cIdx = 0; cIdx < nParams; cIdx++)
                  p[cIdx] = x[rowIdx, cIdx];

                fi[fidx] = factor * ((double)(func(p)) - constraints[rowIdx, colIdx]);
                var g = (double[])AD.Grad(func, p);
                for (int jacIdx = 0; jacIdx < c.Length; jacIdx++) {
                  jac[fidx, jacIdx] = factor * g[nParams + jacIdx]; // skip the elements for the variable values
                }
                fidx++;
              } else {
                // df / dxi constraint
                var g = (double[])AD.Grad(func, p);
                fi[fidx] = factor * g[colIdx];
                var hess = AD.Hessian(func, p);
                for (int jacIdx = 0; jacIdx < c.Length; jacIdx++) {
                  jac[fidx, jacIdx] = factor * (double)hess[nParams + colIdx, nParams + jacIdx]; // skip the elements for the variable values
                }
                fidx++;
              }
            }
          }
        }
      };
    }

    public static bool CanOptimizeConstants(ISymbolicExpressionTree tree) {
      return TreeToAutoDiffTermConverter.IsCompatible(tree);
    }
  }
}
